import json

import pytest
from aioresponses import aioresponses

from ta_cmi.api import API, ApiError, InvalidCredentialsError


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


TEST_URL = "http://localhost"

DUMMY_RESPONSE = {"message": "Ok"}
DUMMY_BODY = {"data": "value"}

api = API()


@pytest.mark.asyncio
async def test_make_request_no_json_successful_request_get(mock_aioresponse):
    """Test the _make_request_no_json method with a valid response and get method."""
    mock_aioresponse.get(TEST_URL, status=200, payload=DUMMY_RESPONSE)

    result = await api._make_request_no_json(TEST_URL)

    assert result == json.dumps(DUMMY_RESPONSE)


@pytest.mark.asyncio
async def test_make_request_no_json_successful_request_post(mock_aioresponse):
    """Test the _make_request_no_json method with a valid response and post method."""
    mock_aioresponse.post(TEST_URL, status=200, payload=DUMMY_RESPONSE)

    result = await api._make_request_no_json(TEST_URL, method="POST")

    assert result == json.dumps(DUMMY_RESPONSE)


@pytest.mark.asyncio
async def test_make_request_no_json_successful_empty_request(mock_aioresponse):
    """Test the _make_request_no_json method with a valid empty response."""
    mock_aioresponse.get(TEST_URL, status=204)

    result = await api._make_request_no_json(TEST_URL)

    assert len(result) == 0


@pytest.mark.asyncio
async def test_make_request_no_json_unauthorized_request(mock_aioresponse):
    """Test the _make_request_no_json method with an unauthorized response."""
    mock_aioresponse.get(TEST_URL, status=401)

    with pytest.raises(InvalidCredentialsError) as exc_info:
        await api._make_request_no_json(TEST_URL)

    assert str(exc_info.value) == "Invalid credentials"


@pytest.mark.asyncio
async def test_make_request_no_json_invalid_request(mock_aioresponse):
    """Test the _make_request_no_json method with an invalid response."""
    mock_aioresponse.get(TEST_URL, status=404)

    with pytest.raises(ApiError) as exc_info:
        await api._make_request_no_json(TEST_URL)

    assert str(exc_info.value) == "Invalid response from localhost: 404"


@pytest.mark.asyncio
async def test_make_request_get_json_conversion(mock_aioresponse):
    """Test the _make_request_get method with a raw json response."""
    mock_aioresponse.get(TEST_URL, status=200, payload=DUMMY_RESPONSE)

    result = await api._make_request_get(TEST_URL)

    assert result == DUMMY_RESPONSE


@pytest.mark.asyncio
async def test_make_request_get_json_successful_empty_request(mock_aioresponse):
    """Test the _make_request_get method with a valid empty response."""
    mock_aioresponse.get(TEST_URL, status=204)

    result = await api._make_request_get(TEST_URL)

    assert len(result) == 0


@pytest.mark.asyncio
async def test_make_request_post_json_conversion(mock_aioresponse):
    """Test the _make_request_post method with a raw json response."""
    mock_aioresponse.post(TEST_URL, status=200, payload=DUMMY_RESPONSE)

    result = await api._make_request_post(TEST_URL, DUMMY_BODY)

    mock_aioresponse.assert_called_once_with(
        TEST_URL, method="POST", json=DUMMY_BODY, auth=None
    )

    assert result == DUMMY_RESPONSE


@pytest.mark.asyncio
async def test_make_request_post_json_successful_empty_request(mock_aioresponse):
    """Test the _make_request_post method with a valid empty response."""
    mock_aioresponse.post(TEST_URL, status=204)

    result = await api._make_request_post(TEST_URL, DUMMY_BODY)

    mock_aioresponse.assert_called_once_with(
        TEST_URL, method="POST", json=DUMMY_BODY, auth=None
    )

    assert len(result) == 0



from .channel import Channel
from .const import ChannelMode, ReadOnlyClass


class CoEChannel(Channel, metaclass=ReadOnlyClass):
    """Class to display an digital or analog datapoint."""

    def __init__(self, mode: ChannelMode, index: int, value: float, unit: str) -> None:
        """Initialize and parse json to get properties."""
        super().__init__(None, mode, index, value, unit)



from typing import Any, Dict

from . import Channel
from .const import ChannelType, ReadOnlyClass


class CMIChannel(Channel, metaclass=ReadOnlyClass):
    """Class to display an input or output."""

    def __init__(self, channel_type: ChannelType, json: Dict[str, Any]) -> None:
        """Initialize and parse json to get properties."""
        super().__init__(
            channel_type,
            json["AD"],
            json["Number"],
            json["Value"]["Value"],
            json["Value"]["Unit"],
        )



import math
from unittest.mock import patch

import pytest

from ta_cmi import ChannelMode, ChannelType
from ta_cmi.coe import CoE
from ta_cmi.coe_api import CoEServerConfig
from ta_cmi.coe_channel import CoEChannel
from tests import (
    COE_CONFIG_PACKAGE,
    COE_DATA_PACKAGE,
    COE_SEND_ANALOG_PACKAGE,
    COE_SEND_ANALOG_V2_PACKAGE,
    COE_SEND_DIGITAL_PACKAGE,
    COE_SEND_DIGITAL_V2_PACKAGE,
    COE_VERSION_CHECK_PACKAGE,
    COE_VERSION_PACKAGE,
)

TEST_HOST = "http://localhost"

DATA_RESPONSE = {
    "digital": [{"value": True, "unit": 43}],
    "analog": [{"value": 34.4, "unit": 1}],
    "last_update_unix": 1680410064.03764,
    "last_update": "2023-04-01T12:00:00",
}

DATA_RESPONSE_WITH_EMPTY_SLOTS = {
    "digital": [{"value": True, "unit": 43}, {"value": False, "unit": 0}],
    "analog": [
        {"value": 34.4, "unit": 1},
        {"value": 0.0, "unit": 0},
        {"value": 3.5, "unit": 0},
    ],
    "last_update_unix": 1680410064.03764,
    "last_update": "2023-04-01T12:00:00",
}


@pytest.mark.asyncio
async def test_coe_check_version_first_call():
    """Test the check_version on the first call."""
    coe = CoE(TEST_HOST)

    with patch(
        COE_VERSION_PACKAGE, return_value=coe._min_required_server_version_
    ) as version_mock:
        result = await coe.check_version()

        assert result

        assert coe._version_check

        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_coe_check_version_first_call_invalid_response():
    """Test the check_version on the first call with an invalid response."""
    coe = CoE(TEST_HOST)

    with patch(COE_VERSION_PACKAGE, return_value=None) as version_mock:
        result = await coe.check_version()

        assert result == False

        assert coe._version_check is None

        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_coe_check_version_first_call_invalid_version():
    """Test the check_version on the first call with an invalid version."""
    coe = CoE(TEST_HOST)

    with patch(COE_VERSION_PACKAGE, return_value="2.0.0") as version_mock:
        result = await coe.check_version()

        assert result == False

        assert coe._version_check == False

        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_coe_check_version_second_call():
    """Test the check_version on the second call."""
    coe = CoE(TEST_HOST)

    coe._version_check = True

    with patch(COE_VERSION_PACKAGE) as version_mock:
        result = await coe.check_version()

        assert result

        version_mock.assert_not_called()


@pytest.mark.asyncio
async def test_coe_check_version_second_call_invalid_version():
    """Test the check_version on the second call with an invalid version."""
    coe = CoE(TEST_HOST)

    coe._version_check = False

    with patch(COE_VERSION_PACKAGE) as version_mock:
        result = await coe.check_version()

        assert result == False

        version_mock.assert_not_called()


@pytest.mark.asyncio
async def test_update_no_data_received():
    """Test the update method with empty data."""
    coe = CoE(TEST_HOST)

    coe._channels = {}

    with patch(COE_DATA_PACKAGE, return_value=None) as data_mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.update(1)

        assert coe._channels == {}

        data_mock.assert_called_once()
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_update_old_data_received():
    """Test the update method with old data."""
    coe = CoE(TEST_HOST)

    can_id = 42

    coe._channels[can_id] = {ChannelType.INPUT: {}}
    coe.last_update[can_id] = DATA_RESPONSE["last_update_unix"]

    with patch(COE_DATA_PACKAGE, return_value=DATA_RESPONSE) as data_mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.update(can_id)

        assert coe._channels[can_id] == {ChannelType.INPUT: {}}

        data_mock.assert_called_once()
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_update_data_received():
    """Test the update method with new data."""
    coe = CoE(TEST_HOST)

    can_id = 42

    assert coe._channels == {}
    assert coe.last_update == {}

    with patch(COE_DATA_PACKAGE, return_value=DATA_RESPONSE) as data_mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.update(can_id)

        assert coe._channels[can_id] == {
            ChannelMode.DIGITAL: {1: CoEChannel(ChannelMode.DIGITAL, 1, 1, "43")},
            ChannelMode.ANALOG: {1: CoEChannel(ChannelMode.ANALOG, 1, 34.4, "1")},
        }

        assert math.isclose(coe.last_update[can_id], 1680410064.03764)

        data_mock.assert_called_once()
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_update_data_received_multiple_ids():
    """Test the update method with new data with an existing other id."""
    coe = CoE(TEST_HOST)

    can_id = 42
    can_id2 = 31

    coe._channels[can_id2] = {
        ChannelMode.DIGITAL: {12: CoEChannel(ChannelMode.DIGITAL, 1, 1, "43")}
    }

    coe.last_update = {can_id: 0, can_id2: 1111111}

    assert coe._channels == {
        can_id2: {
            ChannelMode.DIGITAL: {12: CoEChannel(ChannelMode.DIGITAL, 1, 1, "43")}
        }
    }

    assert coe.last_update[can_id] == 0
    assert coe.last_update[can_id2] == 1111111

    with patch(COE_DATA_PACKAGE, return_value=DATA_RESPONSE) as data_mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.update(can_id)

        assert coe._channels[can_id] == {
            ChannelMode.DIGITAL: {1: CoEChannel(ChannelMode.DIGITAL, 1, 1, "43")},
            ChannelMode.ANALOG: {1: CoEChannel(ChannelMode.ANALOG, 1, 34.4, "1")},
        }

        assert coe._channels[can_id2] == {
            ChannelMode.DIGITAL: {12: CoEChannel(ChannelMode.DIGITAL, 1, 1, "43")}
        }

        assert math.isclose(coe.last_update[can_id], 1680410064.03764)

        assert math.isclose(coe.last_update[can_id2], 1111111)

        data_mock.assert_called_once()
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_update_data_received_with_empty_slots():
    """Test the update method with new data with empty slots."""
    coe = CoE(TEST_HOST)

    can_id = 42

    assert coe._channels == {}

    with patch(
        COE_DATA_PACKAGE, return_value=DATA_RESPONSE_WITH_EMPTY_SLOTS
    ) as data_mock, patch(COE_VERSION_CHECK_PACKAGE) as version_mock:
        await coe.update(can_id)

        assert coe._channels[can_id] == {
            ChannelMode.DIGITAL: {1: CoEChannel(ChannelMode.DIGITAL, 1, 1, "43")},
            ChannelMode.ANALOG: {
                1: CoEChannel(ChannelMode.ANALOG, 1, 34.4, "1"),
                3: CoEChannel(ChannelMode.ANALOG, 3, 3.5, "0"),
            },
        }

        data_mock.assert_called_once()
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_get_server_version():
    """Test the test_get_server_version."""
    coe = CoE(TEST_HOST)

    version = "1.1.0"

    with patch(COE_VERSION_PACKAGE, return_value=version) as version_mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_check_mock:
        result = await coe.get_server_version()

        assert result == version

        version_mock.assert_called_once()
        version_check_mock.assert_called_once()


@pytest.mark.asyncio
async def test_get_server_config():
    """Test the get_server_config."""
    coe = CoE(TEST_HOST)

    config = CoEServerConfig(coe_version=2)

    with patch(COE_CONFIG_PACKAGE, return_value=config) as config_mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_check_mock:
        result = await coe.get_server_config()

        assert result == config

        config_mock.assert_called_once()
        version_check_mock.assert_called_once()


@pytest.mark.asyncio
async def test_send_analog_values():
    """Test the send_analog_values."""
    coe = CoE(TEST_HOST)

    data = [CoEChannel(ChannelMode.ANALOG, 0, 1, "43")]
    page = 1

    with patch(COE_SEND_ANALOG_PACKAGE) as mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.send_analog_values(data, page)

        mock.assert_called_once_with(data, page)
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_send_digital_values():
    """Test the send_digital_values."""
    coe = CoE(TEST_HOST)

    data = [CoEChannel(ChannelMode.DIGITAL, 0, 1, "")]
    page = False

    with patch(COE_SEND_DIGITAL_PACKAGE) as mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.send_digital_values(data, page)

        mock.assert_called_once_with(data, page)
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_send_analog_values_v2():
    """Test the send_analog_values V2."""
    coe = CoE(TEST_HOST)

    data = [CoEChannel(ChannelMode.ANALOG, 1, 1, "43")]

    with patch(COE_SEND_ANALOG_V2_PACKAGE) as mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.send_analog_values_v2(data)

        mock.assert_called_once_with(data)
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_send_digital_values_v2():
    """Test the send_digital_values V2."""
    coe = CoE(TEST_HOST)

    data = [CoEChannel(ChannelMode.DIGITAL, 1, 1, "")]

    with patch(COE_SEND_DIGITAL_V2_PACKAGE) as mock, patch(
        COE_VERSION_CHECK_PACKAGE
    ) as version_mock:
        await coe.send_digital_values_v2(data)

        mock.assert_called_once_with(data)
        version_mock.assert_called_once()


@pytest.mark.asyncio
async def test_check_version():
    """Test if the version check fetches the server version."""
    coe = CoE(TEST_HOST)

    with patch(COE_VERSION_PACKAGE, return_value="0.0.0") as version_mock:
        await coe.check_version()

        version_mock.assert_called_once()



from setuptools import setup

with open("README.md", "r", encoding="utf8") as f:
    readme = f.read()

with open("LICENSE", "r", encoding="utf8") as f:
    license = f.read()

setup(
    name="ta-cmi",
    version="3.3.0",
    description="A Python wrapper to read out  sensors from Technische Alternative using the C.M.I.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/DeerMaximum/ta-cmi",
    author="DeerMaximum",
    author_email="2629822-DeerMaximum@users.noreply.gitlab.com",
    license=license,
    packages=["ta_cmi"],
    install_requires=["aiohttp>=3.9.3", "packaging>=24.0"],
)



from typing import Any, Dict, List

from aiohttp import ClientSession
from packaging import version

from .coe_api import CoEAPI, CoEServerConfig
from .coe_channel import CoEChannel
from .const import _LOGGER, ChannelMode


class CoE:
    _min_required_server_version_ = "2.1.0"
    _version_check: bool | None = None

    def __init__(self, host: str, session: ClientSession = None):
        """Initialize."""
        self._channels: Dict[int, Dict[ChannelMode, Dict[int, CoEChannel]]] = {}

        self._api = CoEAPI(host, session)
        self.last_update: Dict[int, float] = {}

    async def check_version(self) -> bool:
        """Check if the library has support for the server."""
        if self._version_check is None:
            ver = await self._api.get_coe_version()

            if ver is None:
                return False

            server_version = version.parse(ver)
            required_version = version.parse(self._min_required_server_version_)

            self._version_check = (
                required_version.major <= server_version.major
                and required_version.minor <= server_version.minor
            )

        if not self._version_check:
            _LOGGER.warning(
                f"This version of the library requires at least the CoE server with version: {self._min_required_server_version_}."
            )

        return self._version_check

    @staticmethod
    def _extract_channels(
        mode: ChannelMode, raw_channels: List[Dict[str, Any]]
    ) -> Dict[int, CoEChannel]:
        """Extract channel info from data array from request."""
        channels: Dict[int, CoEChannel] = {}

        for index, channel_raw in enumerate(raw_channels):
            if channel_raw["unit"] == 0 and channel_raw["value"] == 0:
                continue

            channels[index + 1] = CoEChannel(
                mode, index + 1, float(channel_raw["value"]), str(channel_raw["unit"])
            )

        return channels

    async def update(self, can_id: int) -> None:
        """Update data."""
        await self.check_version()
        _LOGGER.debug("Update CoE data")

        data = await self._api.get_coe_data(can_id)

        if data is None:
            _LOGGER.debug("Received no data from CoE")
            return

        if data["last_update_unix"] <= self.last_update.get(can_id, 0):
            _LOGGER.debug("Received old data from CoE")
            return

        if self._channels.get(can_id, None) is None:
            self._channels[can_id] = {}

        self._channels[can_id][ChannelMode.DIGITAL] = self._extract_channels(
            ChannelMode.DIGITAL, data["digital"]
        )

        self._channels[can_id][ChannelMode.ANALOG] = self._extract_channels(
            ChannelMode.ANALOG, data["analog"]
        )

        self.last_update[can_id] = data["last_update_unix"]

    def get_channels(
        self, can_id: int, channel_mode: ChannelMode
    ) -> Dict[int, CoEChannel]:
        """Get all the fetched channels from a type."""
        return self._channels.get(can_id, {}).get(channel_mode, {})

    async def get_server_version(self) -> str:
        """Get the server version."""
        await self.check_version()
        return await self._api.get_coe_version()

    async def get_server_config(self) -> CoEServerConfig:
        """Get the coe server config."""
        await self.check_version()
        return await self._api.get_coe_server_config()

    async def send_analog_values(self, data: list[CoEChannel], page: int):
        """Send analog values to CoE server."""
        await self.check_version()
        await self._api.send_analog_values(data, page)

    async def send_digital_values(self, data: list[CoEChannel], second_page: bool):
        """Send digital values to CoE server."""
        await self.check_version()
        await self._api.send_digital_values(data, second_page)

    async def send_analog_values_v2(self, channel: list[CoEChannel]):
        """Send a analog values to CoE server (V2)."""
        await self.check_version()
        await self._api.send_analog_values_v2(channel)

    async def send_digital_values_v2(self, channel: list[CoEChannel]):
        """Send a digital values to CoE server (V2)."""
        await self.check_version()
        await self._api.send_digital_values_v2(channel)



from .const import UNITS_DE, UNITS_EN, ChannelMode, ChannelType, Languages


class Channel:
    """Class to display a data point"""

    def __init__(
        self,
        channel_type: ChannelType | None,
        mode: ChannelMode,
        index: int,
        value: float,
        unit: str,
    ) -> None:
        """Initialize and parse json to get properties."""
        self.type = channel_type
        self.mode = mode
        self.index = index
        self.value = value
        self.unit = unit

    def get_unit(self, language: Languages = Languages.EN) -> str:
        """Get the unit of the channel."""
        if language == Languages.EN:
            return UNITS_EN.get(self.unit, "Unknown")
        else:
            return UNITS_DE.get(self.unit, "Unknown")

    def __eq__(self, other) -> bool:
        return (
            self.type == other.type
            and self.mode == other.mode
            and self.index == other.index
            and self.value == other.value
            and self.unit == other.unit
        )

    def __repr__(self) -> str:
        return f"Channel {self.index}: Type: {self.type}, Mode: {self.mode}, Value: {self.value} {self.get_unit()}"



__version__ = "3.3.0"
__author__ = "DeerMaximum"

from .api import API, ApiError, InvalidCredentialsError
from .channel import Channel
from .cmi import CMI
from .cmi_api import CMIAPI, RateLimitError
from .cmi_channel import CMIChannel
from .coe import CoE
from .coe_api import CoEAPI, CoEServerConfig
from .coe_channel import CoEChannel
from .const import ChannelMode, ChannelType, Languages
from .device import Device, InvalidDeviceError



from unittest.mock import patch

import pytest

from ta_cmi import ChannelMode, CoEChannel
from ta_cmi.coe_api import CoEAPI, CoEServerConfig

from . import MAKE_GET_REQUEST_PACKAGE, MAKE_POST_REQUEST_PACKAGE

TEST_HOST = "http://localhost"

DATA_RESPONSE = {
    "digital": [{"value": True, "unit": 43}],
    "analog": [{"value": 34.4, "unit": 1}],
    "last_update_unix": 1680410064.03764,
    "last_update": "2023-04-01T12:00:00",
}

VERSION_RESPONSE = {"version": "1.1.0"}

CONFIG_RESPONSE = {"coe_version": 2}

api = CoEAPI(TEST_HOST)


@pytest.mark.asyncio
async def test_get_coe_data_with_data_response():
    """Test the get_coe_data with a valid data response."""
    can_id = 42

    with patch(MAKE_GET_REQUEST_PACKAGE, return_value=DATA_RESPONSE) as request_mock:
        result = await api.get_coe_data(can_id)

        assert result == DATA_RESPONSE

        request_mock.assert_called_once_with(f"{TEST_HOST}/receive/{can_id}")


@pytest.mark.asyncio
async def test_get_coe_data_with_empty_response():
    """Test the get_coe_data with a valid empty response."""
    can_id = 42

    with patch(MAKE_GET_REQUEST_PACKAGE, return_value={}) as request_mock:
        result = await api.get_coe_data(can_id)

        assert result is None

        request_mock.assert_called_once_with(f"{TEST_HOST}/receive/{can_id}")


@pytest.mark.asyncio
async def test_get_server_version_with_data_response():
    """Test the get_coe_version with a valid data response."""
    with patch(MAKE_GET_REQUEST_PACKAGE, return_value=VERSION_RESPONSE) as request_mock:
        result = await api.get_coe_version()

        assert result == VERSION_RESPONSE["version"]

        request_mock.assert_called_once()


@pytest.mark.asyncio
async def test_get_server_version_with_empty_response():
    """Test the get_coe_version with a valid empty response."""
    with patch(MAKE_GET_REQUEST_PACKAGE, return_value={}) as request_mock:
        result = await api.get_coe_version()

        assert result is None

        request_mock.assert_called_once()


@pytest.mark.asyncio
async def test_get_server_version():
    """Test the get_coe_version response."""
    with patch(MAKE_GET_REQUEST_PACKAGE, return_value=CONFIG_RESPONSE) as request_mock:
        result = await api.get_coe_server_config()

        server_config = CoEServerConfig(**CONFIG_RESPONSE)

        assert result == server_config

        request_mock.assert_called_once()


def test_check_channel_mode_valid_channels():
    """Test the check_channel_mode with only valid channels."""
    test_data = [
        CoEChannel(ChannelMode.DIGITAL, 0, 0, ""),
        CoEChannel(ChannelMode.DIGITAL, 0, 0, ""),
    ]

    result = api._check_channel_mode(ChannelMode.DIGITAL, test_data)

    assert result


def test_check_channel_mode_invalid_channels():
    """Test the check_channel_mode with invalid channels."""
    test_data = [
        CoEChannel(ChannelMode.ANALOG, 0, 0, ""),
        CoEChannel(ChannelMode.DIGITAL, 0, 0, ""),
    ]

    result = api._check_channel_mode(ChannelMode.ANALOG, test_data)

    assert result == False


@pytest.mark.asyncio
async def test_send_digital_value():
    """Test the send_digital_value ."""

    expected_data = {
        "values": [
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ],
        "page": 1,
    }

    test_data = [CoEChannel(ChannelMode.DIGITAL, 0, 1, "")]

    test_data = test_data + [
        CoEChannel(ChannelMode.DIGITAL, 0, 0, "") for _ in range(15)
    ]

    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        await api.send_digital_values(test_data, True)

        request_mock.assert_called_once_with(f"{TEST_HOST}/send/digital", expected_data)


@pytest.mark.asyncio
async def test_send_analog_value():
    """Test the send_digital_value ."""

    expected_data = {
        "values": [
            {"value": 15.5, "unit": 46},
            {"value": 0, "unit": 0},
            {"value": 15.5, "unit": 46},
            {"value": 0, "unit": 0},
        ],
        "page": 1,
    }

    test_data = [
        CoEChannel(ChannelMode.ANALOG, 0, 15.5, "46"),
        CoEChannel(ChannelMode.ANALOG, 0, 0, "0"),
        CoEChannel(ChannelMode.ANALOG, 0, 15.5, "46"),
        CoEChannel(ChannelMode.ANALOG, 0, 0, "0"),
    ]

    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        await api.send_analog_values(test_data, 1)

        request_mock.assert_called_once_with(f"{TEST_HOST}/send/analog", expected_data)


@pytest.mark.asyncio
async def test_send_analog_value_v2_invalid_channel():
    """Test the send_analog_value_v2 with an invalid channel."""
    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        test_channel = CoEChannel(ChannelMode.DIGITAL, 1, 1, "46")
        await api.send_analog_values_v2([test_channel])

        request_mock.assert_not_called()


@pytest.mark.asyncio
async def test_send_analog_value_v2_invalid_channel_index():
    """Test the send_analog_value_v2 with an invalid channel index."""
    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        test_channel = CoEChannel(ChannelMode.ANALOG, 0, 1, "46")
        await api.send_analog_values_v2([test_channel])

        request_mock.assert_not_called()


@pytest.mark.asyncio
async def test_send_analog_value_v2_valid():
    """Test the send_analog_value_v2 with valid data."""
    expected_data = {"values": [{"channel": 1, "value": 16.4, "unit": "46"}]}

    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        test_channel = CoEChannel(ChannelMode.ANALOG, 1, 16.4, "46")
        await api.send_analog_values_v2([test_channel])

        request_mock.assert_called_once_with(
            f"{TEST_HOST}/send/v2/analog", expected_data
        )


@pytest.mark.asyncio
async def test_send_digital_value_v2_invalid_channel():
    """Test the send_digital_value_v2 with an invalid channel."""
    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        test_channel = CoEChannel(ChannelMode.ANALOG, 1, 1, "46")
        await api.send_digital_values_v2([test_channel])

        request_mock.assert_not_called()


@pytest.mark.asyncio
async def test_send_digital_value_v2_invalid_channel_index():
    """Test the send_digital_value_v2 with an invalid channel index."""
    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        test_channel = CoEChannel(ChannelMode.DIGITAL, 0, 1, "46")
        await api.send_digital_values_v2([test_channel])

        request_mock.assert_not_called()


@pytest.mark.asyncio
async def test_send_digital_value_v2_valid():
    """Test the send_digital_value_v2 with valid data."""
    expected_data = {"values": [{"channel": 1, "value": 0, "unit": "46"}]}

    with patch(MAKE_POST_REQUEST_PACKAGE) as request_mock:
        test_channel = CoEChannel(ChannelMode.DIGITAL, 1, 0, "46")
        await api.send_digital_values_v2([test_channel])

        request_mock.assert_called_once_with(
            f"{TEST_HOST}/send/v2/digital", expected_data
        )



from unittest.mock import patch

import pytest

from ta_cmi import CMI

TEST_HOST = "http://localhost"


@pytest.mark.asyncio
async def test_get_devices_create_all_provided_devices():
    """Fetch create devices that are provided."""
    cmi: CMI = CMI(TEST_HOST, "user", "pass")

    with patch(
        "ta_cmi.cmi_api.CMIAPI.get_devices_ids", return_value=["2", "3", "4"]
    ) as device_mock:
        devices = await cmi.get_devices()

        device_mock.assert_called_once()

        assert len(devices) == 3

        assert devices[0].id == "2"
        assert devices[1].id == "3"
        assert devices[2].id == "4"
